@@ -21,16 +21,16 @@
  * specific prior written permission.
  * 
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
- * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
- * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
- * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
- * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
- * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
- * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
- * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
- * POSSIBILITY OF SUCH DAMAGE.
+ * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
+ * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 /**
@@ -47,13 +47,13 @@
 #include "net_help.h"
 #include "ubhook.h"
 #include "reshook.h"
+#include "http.h"
+#include "update.h"
 #include <ldns/ldns.h>
 
 /* create probes for the ip addresses in the string */
 static void probe_spawn(const char* ip, int recurse, int dnstcp,
 	struct ssllist* ssldns, int port);
-/* delete and stop outq */
-static void outq_delete(struct outq* outq);
 /* set timeout on outq and create UDP query and send it */
 static int outq_settimeout_and_send(struct outq* outq);
 /* send outq over tcp */
@@ -68,6 +68,10 @@ void probe_start(char* ips)
 {
 	char* next;
 	struct svr* svr = global_svr;
+	if(svr->http) {
+		http_general_delete(svr->http);
+		svr->http = NULL;
+	}
 	if(svr->probes) {
 		/* clear existing probe list */
 		probe_list_delete(svr->probes);
@@ -105,6 +109,13 @@ void probe_start(char* ips)
 		verbose(VERB_OPS, "probe started: but still forced insecure");
 		/* call it right away, so the user does not have to wait */
 		probe_setup_hotspot_signon(svr);
+	} else {
+		/* there are cache DNS, and not forced insecure: check HTTP */
+		if(!svr->http && svr->cfg->num_http_urls != 0) {
+			svr->http = http_general_start(svr);
+			if(!svr->http) log_err("out of memory");
+			svr->http->saw_http_work = 0;
+		}
 	}
 }
 
@@ -113,10 +124,13 @@ void probe_delete(struct probe_ip* p)
 	if(!p) return;
 	free(p->name);
 	free(p->reason);
+	free(p->http_desc);
 	SSL_CTX_free(p->sslctx);
 	outq_delete(p->ds_c);
 	outq_delete(p->dnskey_c);
 	outq_delete(p->nsec3_c);
+	outq_delete(p->host_c);
+	http_get_delete(p->http);
 	free(p);
 }
 
@@ -142,7 +156,7 @@ get_random_dest(void)
 static const char*
 get_random_nsec3_dest(void)
 {
-	const char* choices[] = { "com.", "uk.", "nl.", "de." };
+	const char* choices[] = { "_probe.us.com.", "_probe.uk.com.", "_probe.kr.com.", "_probe.uk.net." };
 	return choices[ ldns_get_random() % 4 ];
 }
 
@@ -158,7 +172,7 @@ get_random_auth_ip4(void)
 		"198.41.0.4", /* a */
 		"192.228.79.201", /* b */
 		"192.33.4.12", /* c */
-		"128.8.10.90", /* d */
+		"199.7.91.13", /* d */
 		"192.203.230.10", /* e */
 		"192.5.5.241", /* f */
 		"192.112.36.4", /* g */
@@ -179,6 +193,7 @@ get_random_auth_ip6(void)
 	/* list of root servers */
 	const char* choices[] = {
 		"2001:503:ba3e::2:30", /* a */
+		"2001:500:2::c", /* c */
 		"2001:500:2d::d", /* d */
 		"2001:500:2f::f", /* f */
 		"2001:500:1::803f:235", /* h */
@@ -188,7 +203,7 @@ get_random_auth_ip6(void)
 		"2001:500:3::42", /* l */
 		"2001:dc3::35" /* m */
 	};
-	return choices[ ldns_get_random() % 9 ];
+	return choices[ ldns_get_random() % 10 ];
 }
 
 static const char* get_random_tcp80_ip4(struct cfg* cfg)
@@ -239,6 +254,59 @@ static struct ssllist* get_random_ssl443_ip6(struct cfg* cfg)
 		((unsigned)ldns_get_random())%((unsigned)cfg->num_ssl443_ip6));
 }
 
+int probe_is_cache(struct probe_ip* p)
+{
+	return !p->to_auth && !p->ssldns && !p->dnstcp && !p->to_http;
+}
+
+/** check if hash of x is the given hash, error or NULL on success. */
+static char*
+match_hash(struct hashlist* he, X509* x)
+{
+	unsigned char hash[EVP_MAX_MD_SIZE];
+	unsigned int len = (unsigned int)sizeof(hash);
+	/* always sha256 now */
+	if(!X509_digest(x, EVP_sha256(), hash, &len)) {
+		log_err("out of memory");
+		return "out of memory in X509_digest";
+	}
+	if(verbosity >= 3) {
+		char buf[1024];
+		char* at = buf;
+		size_t blen = sizeof(buf);
+		unsigned int i;
+		for(i=0; i<len; i++) {
+			snprintf(at, blen, "%2.2X%s",
+			    (unsigned int)hash[i], (i==len-1)?"":":");
+			blen -= strlen(at);
+			at += strlen(at); 
+		}
+		log_info("the ssl fingerprint of the server is %s", buf);
+	}
+
+	if(len != he->hashlen) {
+		/* should not happen, since we test sha256 length of
+		 * the cfg hash we read in */
+		return "SSL certificate hash length is wrong";
+	}
+	if(memcmp(hash, he->hash, len) != 0) {
+		return "SSL certificate on internet does not match stored hash (ssl intercepted?)";
+	}
+	return 0;
+}
+
+/** check list of hashes, error string or NULL on success */
+static char*
+match_hashes(struct hashlist* list, X509* x)
+{
+	struct hashlist* h;
+	char* reason = "Internal error: no hashlist but comparison";
+	for(h=list; h; h=h->next)
+		if( (reason=match_hash(h, x)) == NULL)
+			return NULL;
+	return reason;
+}
+
 /** check SSL certificate on probe (that is otherwise DNSSEC OK) */
 static const char*
 check_ssl(struct outq* outq)
@@ -249,40 +317,15 @@ check_ssl(struct outq* outq)
 	 * a stored hash, you can compute this hash with the command
 	 * openssl x509 -sha256 -fingerprint -in server.pem */
 	X509* x = SSL_get_peer_certificate(outq->c->ssl);
-	if(!outq->probe->ssldns->has_hash) {
+	if(!outq->probe->ssldns->hashes) {
 		/* no stored hash */
 		X509_free(x);
 	} else if(x) {
-		unsigned char hash[EVP_MAX_MD_SIZE];
-		unsigned int len = (unsigned int)sizeof(hash);
-		if(!X509_digest(x, EVP_sha256(), hash, &len)) {
-			log_err("out of memory");
-			X509_free(x);
-			return "out of memory in X509_digest";
-		}
+		char* reason = match_hashes(outq->probe->ssldns->hashes, x);
 		X509_free(x);
-		if(verbosity >= 3) {
-			char buf[1024];
-			char* at = buf;
-			size_t blen = sizeof(buf);
-			unsigned int i;
-			for(i=0; i<len; i++) {
-				size_t a = snprintf(at, blen, "%2.2X%s",
-					(int)hash[i], (i==len-1)?"":":");
-				at += a; 
-				blen -= a;
-			}
-			log_info("the ssl fingerprint of server %s is %s",
-				outq->probe->name, buf);
-		}
-		if(len != outq->probe->ssldns->hashlen) {
-			/* should not happen, since we test sha256 length of
-			 * the cfg hash we read in */
-			return "SSL certificate hash length is wrong";
-		}
-		if(memcmp(hash, outq->probe->ssldns->hash, len) != 0) {
-			return "SSL certificate on internet does not match stored hash (ssl intercepted?)";
-		}
+		if(reason) 
+			return reason;
+		
 	} else {
 		return "No SSL certificate on internet but have stored hash (ssl intercepted?)";
 	}
@@ -296,6 +339,10 @@ outq_done(struct outq* outq, const char* reason)
 {
 	struct probe_ip* p = outq->probe;
 	const char* in = NULL;
+	if(!p) {
+		selfupdate_outq_done(global_svr->update, outq, NULL, reason);
+		return;
+	}
 	if(p->sslctx && !reason) {
 		reason = check_ssl(outq);
 	}
@@ -303,6 +350,9 @@ outq_done(struct outq* outq, const char* reason)
 		outq_delete(p->nsec3_c);
 		p->nsec3_c = NULL;
 		in = "NSEC3";
+	} else if(p->host_c == outq) {
+		http_host_outq_done(p, reason);
+		return;
 	} else if(p->ds_c == outq) {
 		outq_delete(p->ds_c);
 		p->ds_c = NULL;
@@ -383,7 +433,18 @@ outq_check_packet(struct outq* outq, uint8_t* wire, size_t len)
 	ldns_pkt *p = NULL;
 	ldns_status s;
 	if(verbosity >= VERB_ALGO) {
-		verbose(VERB_ALGO, "from %s %s %s received",
+		if(!outq->probe) {
+		    verbose(VERB_ALGO, "%s %s received", outq->qname,
+		    	outq->qtype==LDNS_RR_TYPE_TXT?"TXT":(
+		    	outq->qtype==LDNS_RR_TYPE_A?"A":(
+		    	outq->qtype==LDNS_RR_TYPE_AAAA?"AAAA":"other"
+			)));
+		} else if(outq->probe->to_http) {
+		    verbose(VERB_ALGO, "from %s %s received",
+			outq->probe->name, outq->probe->http_ip6?
+			"HTTPip6":"HTTPip4");
+		} else
+		    verbose(VERB_ALGO, "from %s %s %s received",
 			outq->probe->name,
 			outq->qtype==PROBE_NSEC3_QTYPE?"NSEC3":(
 			outq->qtype==LDNS_RR_TYPE_DNSKEY?"DNSKEY":"DS"),
@@ -393,7 +454,8 @@ outq_check_packet(struct outq* outq, uint8_t* wire, size_t len)
 			)));
 		log_hex("packet", wire, len);
 	}
-	outq->probe->got_packet = 1;
+	if(outq->probe)
+		outq->probe->got_packet = 1;
 
 	if(!LDNS_QR_WIRE(wire)) {
 		outq_done(outq, "reply without QR flag");
@@ -402,7 +464,7 @@ outq_check_packet(struct outq* outq, uint8_t* wire, size_t len)
 	if(LDNS_TC_WIRE(wire)) {
 		/* start TCP query and wait for it */
 		verbose(VERB_ALGO, "%s: TC flag, switching to TCP",
-			outq->probe->name);
+			outq->probe?outq->probe->name:outq->qname);
 		if(!outq_send_tcp(outq)) {
 			outq_done(outq, "cannot send TCP query after TC flag");
 		}
@@ -434,6 +496,16 @@ outq_check_packet(struct outq* outq, uint8_t* wire, size_t len)
 		ldns_pkt_free(p);
 		return;
 	}
+	if(!outq->probe) {
+		selfupdate_outq_done(global_svr->update, outq, p, NULL);
+		return;
+	}
+	/* if this all OK, and addr, then use http addr process */
+	if(outq == outq->probe->host_c) {
+		/* this routine frees pkt */
+		http_host_outq_result(outq->probe, p);
+		return;
+	}
 
 	/* test EDNS0 presence, of OPT record */
 	/* LDNS forgets during pkt parse, but we test the ARCOUNT;
@@ -550,17 +622,26 @@ int outq_handle_udp(struct comm_point* c, void* my_arg, int error,
 static int
 create_probe_query(struct outq* outq, ldns_buffer* buffer)
 {
+	uint16_t flags = 0;
 	ldns_pkt* pkt = NULL;
-	ldns_status status = ldns_pkt_query_new_frm_str(&pkt, outq->qname,
-		outq->qtype, LDNS_RR_CLASS_IN,
-		(uint16_t)(outq->recurse?LDNS_RD|LDNS_CD:0));
+	ldns_status status;
+	if(outq->recurse)
+		flags |= LDNS_RD;
+	if(outq->cdflag)
+		flags |= LDNS_CD;
+	status = ldns_pkt_query_new_frm_str(&pkt, outq->qname, outq->qtype,
+		LDNS_RR_CLASS_IN, flags);
 	if(status != LDNS_STATUS_OK) {
 		log_err("could not pkt_query_new %s",
 			ldns_get_errorstr_by_id(status));
 		return 0;
 	}
-	ldns_pkt_set_edns_do(pkt, 1);
-	ldns_pkt_set_edns_udp_size(pkt, 4096);
+	if(outq->edns) {
+		ldns_pkt_set_edns_do(pkt, 1);
+		ldns_pkt_set_edns_udp_size(pkt, 4096);
+	} else {
+		ldns_pkt_set_edns_do(pkt, 0);
+	}
 	ldns_pkt_set_id(pkt, outq->qid);
 	ldns_buffer_clear(buffer);
 	status = ldns_pkt2buffer_wire(buffer, pkt);
@@ -575,9 +656,9 @@ create_probe_query(struct outq* outq, ldns_buffer* buffer)
 	return 1;
 }
 
-static struct outq*
+struct outq*
 outq_create(const char* ip, int tp, const char* domain, int recurse,
-	struct probe_ip* p, int tcp, int onssl, int port)
+	struct probe_ip* p, int tcp, int onssl, int port, int edns, int cdflag)
 {
 	int fd;
 	struct outq* outq = (struct outq*)calloc(1, sizeof(*outq));
@@ -593,6 +674,8 @@ outq_create(const char* ip, int tp, const char* domain, int recurse,
 	outq->recurse = recurse;
 	outq->on_ssl = onssl;
 	outq->port = port;
+	outq->edns = edns;
+	outq->cdflag = cdflag;
 
 	if(!ipstrtoaddr(ip, port, &outq->addr, &outq->addrlen)) {
 		log_err("could not parse ip %s", ip);
@@ -620,6 +703,22 @@ outq_create(const char* ip, int tp, const char* domain, int recurse,
 
 	fd = socket(strchr(ip, ':')?PF_INET6:PF_INET, SOCK_DGRAM, IPPROTO_UDP);
 	if(fd == -1) {
+#ifndef USE_WINSOCK
+		if(errno == EAFNOSUPPORT || errno == EPROTONOSUPPORT) {
+			if(verbosity <= 2) {
+				free(outq);
+				return NULL;
+			}
+		}
+#else
+		if(WSAGetLastError() == WSAEAFNOSUPPORT ||
+			WSAGetLastError() == WSAEPROTONOSUPPORT) {
+			if(verbosity <= 2) {
+				free(outq);
+				return NULL;
+			}
+		}
+#endif
 		log_err("socket %s udp: %s", strchr(ip, ':')?"ip6":"ip4",
 			strerror(errno));
 		free(outq);
@@ -645,7 +744,7 @@ outq_create(const char* ip, int tp, const char* domain, int recurse,
 	return outq;
 }
 
-static void outq_delete(struct outq* outq)
+void outq_delete(struct outq* outq)
 {
 	if(!outq) return;
 	comm_timer_delete(outq->timer);
@@ -683,9 +782,10 @@ static int outq_settimeout_and_send(struct outq* outq)
 void outq_timeout(void* arg)
 {
 	struct outq* outq = (struct outq*)arg;
+	char *t = ldns_rr_type2str(outq->qtype);
 	verbose(VERB_ALGO, "%s %s: UDP timeout after %d msec",
-		outq->probe->name, outq->qtype==PROBE_NSEC3_QTYPE?"NSEC3":(
-		outq->qtype==LDNS_RR_TYPE_DNSKEY?"DNSKEY":"DS"), outq->timeout);
+		outq->probe?outq->probe->name:outq->qname, t, outq->timeout);
+	free(t);
 	if(outq->timeout > QUERY_END_TIMEOUT) {
 		/* too many timeouts */
 		outq_done(outq, "timeout");
@@ -713,8 +813,19 @@ outq_tcp_take_into_use(struct outq* outq)
 		s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
 	if(s == -1) {
 #ifndef USE_WINSOCK
+		if(errno == EAFNOSUPPORT || errno == EPROTONOSUPPORT) {
+			if(verbosity <= 2) {
+				return 0;
+			}
+		}
 		log_err("outgoing tcp: socket: %s", strerror(errno));
 #else
+		if(WSAGetLastError() == WSAEAFNOSUPPORT ||
+			WSAGetLastError() == WSAEPROTONOSUPPORT) {
+			if(verbosity <= 2) {
+				return 0;
+			}
+		}
 		log_err("outgoing tcp: socket: %s",
 			wsa_strerror(WSAGetLastError()));
 #endif
@@ -885,9 +996,9 @@ static void probe_spawn(const char* ip, int recurse, int dnstcp,
 
 	/* send the probe queries and wait for reply */
 	p->dnskey_c = outq_create(p->name, LDNS_RR_TYPE_DNSKEY, ".",
-		recurse, p, dnstcp, ssldns!=0, port);
+		recurse, p, dnstcp, ssldns!=0, port, 1, 1);
 	p->ds_c = outq_create(p->name, LDNS_RR_TYPE_DS, dest, recurse, p,
-		dnstcp, ssldns!=0, port);
+		dnstcp, ssldns!=0, port, 1, 1);
 	if(!p->ds_c || !p->dnskey_c) {
 		log_err("could not send queries for probe");
 		probe_delete(p);
@@ -898,7 +1009,7 @@ static void probe_spawn(const char* ip, int recurse, int dnstcp,
 		const char* nd = get_random_nsec3_dest();
 		verbose(VERB_ALGO, "nsec3 query %s", nd);
 		p->nsec3_c = outq_create(p->name, PROBE_NSEC3_QTYPE, nd,
-			recurse, p, dnstcp, ssldns!=0, port);
+			recurse, p, dnstcp, ssldns!=0, port, 1, 1);
 		if(!p->nsec3_c) {
 			log_err("could not send nsec3 query for probe");
 			probe_delete(p);
@@ -950,8 +1061,12 @@ static void probe_spawn_ssldns(void)
 
 void probe_unsafe_test(void)
 {
+	int nurl = global_svr->cfg->num_http_urls;
 	verbose(VERB_OPS, "test unsafe probe combination started");
+	/* pretend there is no http probing */
+	global_svr->cfg->num_http_urls = 0;
 	probe_start("127.0.0.3");
+	global_svr->cfg->num_http_urls = nurl;
 	/* pretend we got a reply from the cache (so not disconnected) */
 	if(global_svr->probes)
 		global_svr->probes->got_packet = 1;
@@ -966,8 +1081,12 @@ void probe_unsafe_test(void)
 
 void probe_tcp_test(void)
 {
+	int nurl = global_svr->cfg->num_http_urls;
 	verbose(VERB_OPS, "test tcp probe combination started");
+	/* pretend there is no http probing */
+	global_svr->cfg->num_http_urls = 0;
 	probe_start("127.0.0.3");
+	global_svr->cfg->num_http_urls = nurl;
 	global_svr->probe_direct = 1;
 	probe_spawn("127.0.0.4", 0, 0, 0, DNS_PORT);
 	global_svr->tcp_timer_used = 1; /* avoid retry after 20 sec */
@@ -975,8 +1094,12 @@ void probe_tcp_test(void)
 
 void probe_ssl_test(void)
 {
+	int nurl = global_svr->cfg->num_http_urls;
 	verbose(VERB_OPS, "test ssl probe combination started");
+	/* pretend there is no http probing */
+	global_svr->cfg->num_http_urls = 0;
 	probe_start("127.0.0.3");
+	global_svr->cfg->num_http_urls = nurl;
 	global_svr->probe_direct = 1;
 	probe_spawn("127.0.0.4", 0, 0, 0, DNS_PORT);
 	global_svr->probe_dnstcp = 1;
@@ -984,6 +1107,30 @@ void probe_ssl_test(void)
 	global_svr->tcp_timer_used = 1; /* avoid retry after 20 sec */
 }
 
+void probe_http_test(void)
+{
+	struct cfg* cfg = global_svr->cfg;
+	struct strlist2* old_http_urls = cfg->http_urls;
+	struct strlist2* old_http_urls_last = cfg->http_urls_last;
+	int old_num_http_urls = cfg->num_http_urls;
+	struct strlist2 fake;
+	verbose(VERB_OPS, "test http probe url started");
+	fake.next = NULL;
+	/* lookup of existing URL but we give wrong contents */
+	fake.str1 = "http://open.nlnetlabs.nl/index.html";
+	fake.str2 = "NoSuchString12345";
+	cfg->http_urls = &fake;
+	cfg->http_urls_last = &fake;
+	cfg->num_http_urls = 1;
+	global_svr->skip_http = 0;
+
+	cmd_reprobe();
+
+	cfg->http_urls = old_http_urls;
+	cfg->http_urls_last = old_http_urls_last;
+	cfg->num_http_urls = old_num_http_urls;
+}
+
 /* stop unfininished probes and remove them */
 static void stop_unfinished_probes(void)
 {
@@ -1072,7 +1219,12 @@ probe_done(struct probe_ip* p)
 		if(!svr->probe_dnstcp && !svr->probe_direct &&
 			!svr->saw_first_working) {
 			svr->saw_first_working = 1;
-			if(!svr->forced_insecure) {
+			/* if works, not forced_insecure and http works (or
+			 * did not get probed because not configured) then
+			 * we can already use this cache-DNS now before all
+			 * probes are done */
+			if(!svr->forced_insecure && (!svr->http ||
+				svr->skip_http || svr->http->saw_http_work)) {
 				probe_setup_cache(svr, p);
 			}
 		} else if(svr->probe_direct && !svr->probe_dnstcp &&
@@ -1110,6 +1262,7 @@ void probe_setup_cache(struct svr* svr, struct probe_ip* p)
 	hook_resolv_localhost(svr->cfg);
 	svr_retry_timer_stop();
 	svr->tcp_timer_used = 0;
+	svr->http_insecure = 0;
 }
 
 /** setup for auth (direct to authorities) */
@@ -1123,6 +1276,7 @@ void probe_setup_auth(struct svr* svr)
 	hook_resolv_localhost(svr->cfg);
 	svr_retry_timer_stop();
 	svr->tcp_timer_used = 0;
+	svr->http_insecure = 0;
 }
 
 /** setup for dnstcp (uses tcp to open resolver) */
@@ -1151,6 +1305,7 @@ void probe_setup_dnstcp(struct svr* svr)
 	hook_resolv_localhost(svr->cfg);
 	svr_retry_timer_stop();
 	svr_tcp_timer_enable();
+	svr->http_insecure = 0;
 }
 
 /** setup to be disconnected */
@@ -1163,8 +1318,9 @@ void probe_setup_disconnected(struct svr* svr)
 	/* set resolver.conf to 127.0.0.1 (get rid of old
 	 * settings that may be in there) */
 	hook_resolv_localhost(svr->cfg);
-	svr_retry_timer_next();
+	svr_retry_timer_next(0);
 	svr->tcp_timer_used = 0;
+	svr->http_insecure = 0;
 }
 
 /** setup for dark (no dnssec) */
@@ -1184,8 +1340,9 @@ void probe_setup_dark(struct svr* svr)
 		* the user may select insecure later */
 		hook_resolv_localhost(svr->cfg);
 	}
-	svr_retry_timer_next();
+	svr_retry_timer_next(0);
 	svr->tcp_timer_used = 0;
+	svr->http_insecure = 0;
 }
 
 /** setup forced insecure (for hotspot signon) */
@@ -1199,12 +1356,41 @@ void probe_setup_hotspot_signon(struct svr* svr)
 	hook_resolv_iplist(svr->cfg, svr->probes);
 	svr_retry_timer_stop();
 	svr_tcp_timer_stop();
+	svr->http_insecure = 0;
+}
+
+/** setup http insecure (for hotspot signon) */
+void probe_setup_http_insecure(struct svr* svr)
+{
+	svr->res_state = res_dark;
+	svr->http_insecure = 1;
+	/* effectuate it */
+	hook_unbound_dark(svr->cfg);
+	/* see what the user wants */
+	if(svr->insecure_state) {
+		/* set resolv.conf to DHCP IP list */
+		hook_resolv_iplist(svr->cfg, svr->probes);
+	} else { /* set resolv.conf to 127.0.0.1 now,
+		* the user may select insecure later */
+		hook_resolv_localhost(svr->cfg);
+	}
+	/* set timer to catch user as things work again */
+	svr_retry_timer_next(1);
+	svr->tcp_timer_used = 0;
 }
 
 void
 probe_cache_done(void)
 {
 	struct svr* svr = global_svr;
+	if(!svr->probe_direct && svr->http && !svr->http->saw_http_work
+		&& !svr->skip_http) {
+		/* cache probe completed, but http fails to work, stop probes */
+		/* do not probe direct authority servers, because HTTP fails*/
+		/* unless we skip http probe, then go on to probe authority */
+		probe_all_done();
+		return;
+	}
 	if(!svr->probe_direct && !svr->saw_first_working) {
 		/* no working server, probe the direct DNS */
 		/* we wait until the other probes fail to not put
@@ -1268,7 +1454,11 @@ probe_all_done(void)
 	if(verbosity >= VERB_DETAIL) {
 		struct probe_ip* p;
 		for(p=svr->probes; p; p=p->next) {
-			if(p->dnstcp)
+			if(p->to_http)
+			    verbose(VERB_DETAIL, "%s %s: %s %s",
+			    	p->host_c?"addrlookup":"http", p->name, 
+				p->works?"OK":"error", p->reason?p->reason:"");
+			else if(p->dnstcp)
 			    verbose(VERB_DETAIL, "%s%d %s: %s %s", 
 			    	p->ssldns?"ssl":"tcp", p->port, p->name,
 				p->works?"OK":"error", p->reason?p->reason:"");
@@ -1278,6 +1468,10 @@ probe_all_done(void)
 				p->works?"OK":"error", p->reason?p->reason:"");
 		}
 	}
+	/* reset skip http once it works */
+	if(svr->skip_http && svr->http && svr->http->saw_http_work)
+		svr->skip_http = 0;
+
 	if(svr->forced_insecure) {
 		verbose(VERB_OPS, "probe done: but still forced insecure");
 		/* call it again, in case DHCP changes while hotspot-signon */
@@ -1305,10 +1499,15 @@ probe_all_done(void)
 			verbose(VERB_OPS, "probe done: DNSSEC fails");
 			probe_setup_dark(svr);
 		}
+	} else if(svr->http && !svr->http->saw_http_work && !svr->skip_http) {
+		/* http probed (so have DHCPDNS), but fails */
+		verbose(VERB_OPS, "probe done: http fails");
+		probe_setup_http_insecure(svr);
 	} else {
 		verbose(VERB_OPS, "probe done: DNSSEC to cache");
 		probe_setup_cache(svr, NULL);
 	}
 	svr->probetime = time(0);
 	svr_send_results(svr);
+	svr_check_update(svr);
 }
