@@ -21,16 +21,16 @@
  * specific prior written permission.
  * 
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
- * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
- * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
- * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
- * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
- * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
- * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
- * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
- * POSSIBILITY OF SUCH DAMAGE.
+ * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
+ * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 /**
@@ -42,10 +42,12 @@
 #include "svr.h"
 #include "cfg.h"
 #include "log.h"
+#include "http.h"
 #include "probe.h"
 #include "netevent.h"
 #include "net_help.h"
 #include "reshook.h"
+#include "update.h"
 #ifdef USE_WINSOCK
 #include "winsock_event.h"
 #endif
@@ -91,6 +93,14 @@ struct svr* svr_create(struct cfg* cfg)
 		svr_delete(svr);
 		return NULL;
 	}
+	if(cfg->check_updates) {
+		svr->update = selfupdate_create(svr, cfg);
+		if(!svr->update) {
+			log_err("out of memory");
+			svr_delete(svr);
+			return NULL;
+		}
+	}
 
 	/* setup SSL_CTX */
 	if(!setup_ssl_ctx(svr)) {
@@ -133,9 +143,11 @@ void svr_delete(struct svr* svr)
 	if(svr->ctx) {
 		SSL_CTX_free(svr->ctx);
 	}
+	selfupdate_delete(svr->update);
 	ldns_buffer_free(svr->udp_buffer);
 	comm_timer_delete(svr->retry_timer);
 	comm_timer_delete(svr->tcp_timer);
+	http_general_delete(svr->http);
 	comm_base_delete(svr->base);
 	free(svr);
 }
@@ -383,9 +395,24 @@ int control_callback(struct comm_point* c, void* arg, int err,
 				comm_point_listen_for_rw(c, 0, 1);
 				return 0;
 			} else if(r2 == SSL_ERROR_SYSCALL) {
-				if(errno != 0)
+				if(ERR_peek_error()) {
+					char errbuf[128];
+					ERR_error_string_n(ERR_get_error(),
+						errbuf, sizeof(errbuf));
+					log_err("ssl_handshake: %s", errbuf);
+				} else if(r == 0) {
+					log_err("ssl_handshake EOF violation");
+				} else if(r == -1) {
+#ifdef USE_WINSOCK
+					log_err("ssl_handshake syscall: "
+						"%s, wsa: %s", strerror(errno),
+						wsa_strerror(WSAGetLastError()));
+#else
 					log_err("ssl_handshake syscall: %s",
 						strerror(errno));
+#endif
+				} else	log_err("ssl_handshake syscall ret %d",
+						r);
 				sslconn_delete(s);
 				return 0;
 			} else {
@@ -441,10 +468,13 @@ int control_callback(struct comm_point* c, void* arg, int err,
 		/* we are done handle it */
 		sslconn_command(s);
 	} else if(s->line_state == persist_read) {
-		if(!sslconn_readline(s))
-			return 0;
-		/* we are done handle it */
-		sslconn_persist_command(s);
+		do {
+			if(!sslconn_readline(s))
+				return 0;
+			/* we are done handle it */
+			sslconn_persist_command(s);
+			/* there may be more to read in the same SSL packet */
+		} while(SSL_pending(s->ssl) != 0);
 	} else if(s->line_state == persist_write) {
 		if(sslconn_checkclose(s))
 			return 0;
@@ -486,9 +516,29 @@ static int sslconn_readline(struct sslconn* sc)
 				comm_point_listen_for_rw(sc->c, 0, 1);
 				return 0;
 			} else if(want == SSL_ERROR_SYSCALL) {
-				if(errno != 0)
+				if(ERR_peek_error()) {
+					char errbuf[128];
+					ERR_error_string_n(ERR_get_error(),
+						errbuf, sizeof(errbuf));
+					log_err("ssl_read: %s", errbuf);
+				} else if(r == 0) {
+					log_err("ssl_read EOF violation");
+				} else if(r == -1) {
+#ifdef USE_WINSOCK
+					int wsar = WSAGetLastError();
+					/* conn reset common at restarts */
+					if(wsar == WSAECONNRESET)
+						verbose(VERB_ALGO,
+							"ssl_read syscall: %s",
+							wsa_strerror(wsar));
+					else log_err("ssl_read syscall: "
+						"%s, wsa: %s", strerror(errno),
+						wsa_strerror(wsar));
+#else
 					log_err("ssl_read syscall: %s",
 						strerror(errno));
+#endif
+				} else	log_err("ssl_read syscall ret %d", r);
 				sslconn_delete(sc);
 				return 0;
 			}
@@ -532,10 +582,23 @@ static int sslconn_write(struct sslconn* sc)
 			} else if(want == SSL_ERROR_WANT_WRITE) {
 				return 0;
 			} else if(want == SSL_ERROR_SYSCALL) {
-				/* want syscall, errno==0 : unclean closed */
-				if(errno != 0)
+				if(ERR_peek_error()) {
+					char errbuf[128];
+					ERR_error_string_n(ERR_get_error(),
+						errbuf, sizeof(errbuf));
+					log_err("ssl_write: %s", errbuf);
+				} else if(r == 0) {
+					log_err("ssl_write EOF violation");
+				} else if(r == -1) {
+#ifdef USE_WINSOCK
+					log_err("ssl_write syscall: "
+						"%s, wsa: %s", strerror(errno),
+						wsa_strerror(WSAGetLastError()));
+#else
 					log_err("ssl_write syscall: %s",
 						strerror(errno));
+#endif
+				} else	log_err("ssl_write syscall ret %d", r);
 				sslconn_delete(sc);
 				return 0;
 			}
@@ -593,9 +656,30 @@ static int sslconn_checkclose(struct sslconn* sc)
 			comm_point_listen_for_rw(sc->c, 0, 1);
 			return 0;
 		} else if(want == SSL_ERROR_SYSCALL) {
-			if(errno != 0)
+			if(ERR_peek_error()) {
+				char errbuf[128];
+				ERR_error_string_n(ERR_get_error(),
+					errbuf, sizeof(errbuf));
+				log_err("checkclose ssl_read: %s", errbuf);
+			} else if(r == 0) {
+				log_err("checkclose ssl_read EOF violation");
+			} else if(r == -1) {
+#ifdef USE_WINSOCK
+				int wsar = WSAGetLastError();
+				/* connreset common at restart of system */
+				if(wsar == WSAECONNRESET)
+				    verbose(VERB_ALGO,
+					"checkclose ssl_read syscall: %s",
+					wsa_strerror(wsar));
+				else log_err("checkclose ssl_read syscall: "
+					"%s, wsa: %s", strerror(errno),
+					wsa_strerror(WSAGetLastError()));
+#else
 				log_err("checkclose ssl_read syscall: %s",
 					strerror(errno));
+#endif
+			} else	log_err("checkclose ssl_read syscall ret %d",
+					r);
 			sslconn_delete(sc);
 			return 0;
 		}
@@ -634,7 +718,7 @@ static void persist_cmd_insecure(int val)
 	svr_send_results(svr);
 }
 
-static void cmd_reprobe(void)
+void cmd_reprobe(void)
 {
 	char buf[10240];
 	char* now = buf;
@@ -642,12 +726,13 @@ static void cmd_reprobe(void)
 	struct probe_ip* p;
 	buf[0]=0; /* safe, robust */
 	for(p = global_svr->probes; p; p = p->next) {
-		if(!p->to_auth && !p->dnstcp && !p->ssldns) {
-			int len;
+		if(probe_is_cache(p)) {
+			size_t len;
 			if(left < strlen(p->name)+3)
 				break; /* no space for more */
-			len = snprintf(now, left, "%s%s",
+			snprintf(now, left, "%s%s",
 				(now==buf)?"":" ", p->name);
+			len = strlen(now);
 			left -= len;
 			now += len;
 		}
@@ -662,6 +747,13 @@ static void handle_hotspot_signon_cmd(struct svr* svr)
 	svr_send_results(svr);
 }
 
+static void handle_skip_http_cmd(void)
+{
+	verbose(VERB_OPS, "state skip_http and reprobe");
+	global_svr->skip_http = 1;
+	cmd_reprobe();
+}
+
 static void sslconn_persist_command(struct sslconn* sc)
 {
 	char* str = (char*)ldns_buffer_begin(sc->buffer);
@@ -676,9 +768,16 @@ static void sslconn_persist_command(struct sslconn* sc)
 		persist_cmd_insecure(0);
 	} else if(strcmp(str, "reprobe") == 0) {
 		global_svr->forced_insecure = 0;
+		global_svr->http_insecure = 0;
 		cmd_reprobe();
+	} else if(strcmp(str, "skip_http") == 0) {
+		handle_skip_http_cmd();
 	} else if(strcmp(str, "hotspot_signon") == 0) {
 		handle_hotspot_signon_cmd(global_svr);
+	} else if(strcmp(str, "update_cancel") == 0) {
+		selfupdate_userokay(global_svr->update, 0);
+	} else if(strcmp(str, "update_ok") == 0) {
+		selfupdate_userokay(global_svr->update, 1);
 	} else {
 		log_err("unknown command from panel: %s", str);
 	}
@@ -692,6 +791,14 @@ static void handle_submit(char* ips)
 	probe_start(ips);
 }
 
+/** append update signal to buffer to send */
+static void
+append_update_to_con(struct sslconn* s, char* version_available)
+{
+	ldns_buffer_printf(s->buffer, "update %s\n%s\n\n", PACKAGE_VERSION,
+		version_available);
+}
+
 static void
 send_results_to_con(struct svr* svr, struct sslconn* s)
 {
@@ -705,13 +812,23 @@ send_results_to_con(struct svr* svr, struct sslconn* s)
 		localtime(&svr->probetime)))
 		ldns_buffer_printf(s->buffer, "at %s\n", at);
 	for(p=svr->probes; p; p=p->next) {
-		if(!p->to_auth && !p->dnstcp && !p->ssldns)
+		if(probe_is_cache(p))
 			numcache++;
 		if(!p->finished) {
 			unfinished++;
 			continue;
 		}
-		if(p->dnstcp)
+		if(p->to_http) {
+			if(p->host_c) {
+		    	ldns_buffer_printf(s->buffer, "%s %s %s from %s: %s %s\n",
+		    		"addr", p->host_c->qname,
+				p->http_ip6?"AAAA":"A", p->name,
+				p->works?"OK":"error", p->reason?p->reason:"");
+			} else
+		    	    ldns_buffer_printf(s->buffer, "%s %s (%s): %s %s\n",
+		    		"http", p->http_desc, p->name,
+				p->works?"OK":"error", p->reason?p->reason:"");
+		} else if(p->dnstcp)
 		    ldns_buffer_printf(s->buffer, "%s%d %s: %s %s\n",
 		        p->ssldns?"ssl":"tcp", p->port, p->name,
 			p->works?"OK":"error", p->reason?p->reason:"");
@@ -725,20 +842,46 @@ send_results_to_con(struct svr* svr, struct sslconn* s)
 	else if(!numcache)
 		ldns_buffer_printf(s->buffer, "no cache: no DNS servers have been supplied via DHCP\n");
 
-	ldns_buffer_printf(s->buffer, "state: %s %s\n",
+	ldns_buffer_printf(s->buffer, "state: %s %s%s%s\n",
 		svr->res_state==res_cache?"cache":(
 		svr->res_state==res_tcp?"tcp":(
 		svr->res_state==res_ssl?"ssl":(
 		svr->res_state==res_auth?"auth":(
 		svr->res_state==res_disconn?"disconnected":"nodnssec")))),
-		svr->forced_insecure?"forced_insecure":(
-		svr->insecure_state?"insecure":"secure"));
+		svr->insecure_state?"insecure_mode":"secure",
+		svr->forced_insecure?" forced_insecure":"",
+		svr->http_insecure?" http_insecure":""
+		);
 	ldns_buffer_printf(s->buffer, "\n");
+	if(svr->update && svr->update->update_available &&
+		!svr->update->user_replied) {
+		log_info("append_update signal");
+		append_update_to_con(s, svr->update->version_available);
+	}
 	ldns_buffer_flip(s->buffer);
 	comm_point_listen_for_rw(s->c, 1, 1);
 	s->line_state = persist_write;
 }
 
+void svr_signal_update(struct svr* svr, char* version_available)
+{
+	/* write stop to all connected panels */
+	struct sslconn* s;
+	for(s=svr->busy_list; s; s=s->next) {
+		if(s->line_state == persist_write) {
+			/* busy with last results, fetch update later */
+			s->fetch_another_update=1;
+		}
+		if(s->line_state == persist_write_checkclose) {
+			ldns_buffer_clear(s->buffer);
+			append_update_to_con(s, version_available);
+			ldns_buffer_flip(s->buffer);
+			comm_point_listen_for_rw(s->c, 1, 1);
+			s->line_state = persist_write;
+		}
+	}
+}
+
 static void handle_results_cmd(struct sslconn* sc)
 {
 	/* turn into persist write with results. */
@@ -799,6 +942,19 @@ static void handle_test_ssl_cmd(struct sslconn* sc)
 	sslconn_shutdown(sc);
 }
 
+static void handle_test_http_cmd(struct sslconn* sc)
+{
+	probe_http_test();
+	sslconn_shutdown(sc);
+}
+
+static void handle_test_update_cmd(struct sslconn* sc)
+{
+	global_svr->update->test_flag = 1;
+	global_svr->update_desired = 1;
+	sslconn_shutdown(sc);
+}
+
 static void handle_stoppanels_cmd(struct sslconn* sc)
 {
 	/* write stop to all connected panels */
@@ -810,7 +966,16 @@ static void handle_stoppanels_cmd(struct sslconn* sc)
 			s->line_state != persist_write)
 			continue;
 		(void)SSL_set_mode(s->ssl, SSL_MODE_AUTO_RETRY);
-		fd_set_block(SSL_get_fd(s->ssl));
+		if(SSL_get_fd(s->ssl) != -1) {
+#ifdef USE_WINSOCK
+			/* to be able to set it back to blocking mode
+			 * we have to remove the EventSelect on it */
+			if(WSAEventSelect(SSL_get_fd(s->ssl), NULL, 0)!=0)
+				log_err("WSAEventSelect disable: %s",
+					wsa_strerror(WSAGetLastError()));
+#endif
+			fd_set_block(SSL_get_fd(s->ssl));
+		}
 		if(s->line_state == persist_write) {
 			/* busy with last results,  blocking write them */
 			if(SSL_write(s->ssl, ldns_buffer_current(s->buffer), 
@@ -827,7 +992,8 @@ static void handle_stoppanels_cmd(struct sslconn* sc)
 		}
 
 		/* it will be closed now */
-		fd_set_nonblock(SSL_get_fd(s->ssl));
+		if(SSL_get_fd(s->ssl) != -1)
+			fd_set_nonblock(SSL_get_fd(s->ssl));
 		comm_point_listen_for_rw(s->c, 1, 0);
 		s->line_state = persist_write_checkclose;
 	}
@@ -854,8 +1020,12 @@ static void sslconn_command(struct sslconn* sc)
 		sslconn_shutdown(sc);
 	} else if(strncmp(str, "reprobe", 7) == 0) {
 		global_svr->forced_insecure = 0;
+		global_svr->http_insecure = 0;
 		cmd_reprobe();
 		sslconn_shutdown(sc);
+	} else if(strncmp(str, "skip_http", 9) == 0) {
+		handle_skip_http_cmd();
+		sslconn_shutdown(sc);
 	} else if(strncmp(str, "hotspot_signon", 14) == 0) {
 		handle_hotspot_signon_cmd(global_svr);
 		sslconn_shutdown(sc);
@@ -871,6 +1041,10 @@ static void sslconn_command(struct sslconn* sc)
 		handle_test_tcp_cmd(sc);
 	} else if(strncmp(str, "test_ssl", 8) == 0) {
 		handle_test_ssl_cmd(sc);
+	} else if(strncmp(str, "test_http", 8) == 0) {
+		handle_test_http_cmd(sc);
+	} else if(strncmp(str, "test_update", 11) == 0) {
+		handle_test_update_cmd(sc);
 	} else if(strncmp(str, "stoppanels", 10) == 0) {
 		handle_stoppanels_cmd(sc);
 	} else if(strncmp(str, "stop", 4) == 0) {
@@ -910,29 +1084,38 @@ void svr_retry_callback(void* arg)
 static void svr_retry_setit(struct svr* svr)
 {
 	struct timeval tv;
-	verbose(VERB_ALGO, "retry in %d seconds", svr->retry_timer_timeout);
+	if(svr->retry_timer_count < RETRY_TIMER_COUNT_MAX)
+		verbose(VERB_ALGO, "retry in %d seconds (try nr %d)", svr->retry_timer_timeout, svr->retry_timer_count);
+	else	verbose(VERB_ALGO, "retry in %d seconds", svr->retry_timer_timeout);
 	tv.tv_sec = svr->retry_timer_timeout;
 	tv.tv_usec = 0;
 	comm_timer_set(svr->retry_timer, &tv);
 }
 
-static void svr_retry_start(struct svr* svr)
+static void svr_retry_start(struct svr* svr, int http_mode)
 {
 	svr->retry_timer_timeout = RETRY_TIMER_START;
+	if(http_mode)
+		svr->retry_timer_count = 1;
+	else	svr->retry_timer_count = RETRY_TIMER_COUNT_MAX;
 	svr->retry_timer_enabled = 1;
 	svr_retry_setit(svr);
 }
 
-void svr_retry_timer_next(void)
+void svr_retry_timer_next(int http_mode)
 {
 	struct svr* svr = global_svr;
 	if(!svr->retry_timer_enabled) {
-		svr_retry_start(svr);
+		svr_retry_start(svr, http_mode);
 		return;
 	}
-	svr->retry_timer_timeout *= 2;
-	if(svr->retry_timer_timeout > RETRY_TIMER_MAX)
-		svr->retry_timer_timeout = RETRY_TIMER_MAX;
+	if(svr->retry_timer_count < RETRY_TIMER_COUNT_MAX) {
+		svr->retry_timer_count++;
+	} else {
+		svr->retry_timer_timeout *= 2;
+		if(svr->retry_timer_timeout > RETRY_TIMER_MAX)
+			svr->retry_timer_timeout = RETRY_TIMER_MAX;
+	}
 	svr_retry_setit(svr);
 }
 
@@ -976,3 +1159,12 @@ void svr_tcp_callback(void* arg)
 		cmd_reprobe();
 	}
 }
+
+void svr_check_update(struct svr* svr)
+{
+	if(svr->update_desired && !svr->insecure_state && !svr->forced_insecure
+		&& svr->res_state != res_dark && svr->res_state != res_disconn)
+	{
+		selfupdate_start(svr->update);
+	}
+}
